package org.apache.ojb.broker.metadata.fieldaccess;

/* Copyright 2003-2004 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.Method;

import org.apache.ojb.broker.metadata.MetadataException;
import org.apache.ojb.broker.util.logging.Logger;
import org.apache.ojb.broker.util.logging.LoggerFactory;

/**
 * A {@link PersistentField} implementation using
 * JavaBeans compliant calls only to access persistent attributes.
 * No Reflection is needed. But for each attribute xxx there must be
 * public getXxx() and setXxx() methods.
 *
 * @author <a href="mailto:jbraeuchi@hotmail.com">Jakob Braeuchi</a>
 * @author <a href="mailto:armin@codeAuLait.de">Armin Waibel</a>
 * @version $Id: PersistentFieldIntrospectorImpl.java,v 1.11 2004/06/26 23:51:37 arminw Exp $
 */
public class PersistentFieldIntrospectorImpl extends AbstractPersistentField
{
    private static final long serialVersionUID = 8805309492150404444L;
    private transient PropertyDescriptor propertyDescriptor;

    public PersistentFieldIntrospectorImpl()
    {
        super();
    }

    public PersistentFieldIntrospectorImpl(Class aClass, String aPropertyName)
    {
        super(aClass, aPropertyName);
    }


    /*
    need to overwrite these three methods to avoid
    building of Field instances
    */
    public String getName()
    {
        return this.fieldName;
    }
    public Class getType()
    {
        return getPropertyDescriptor().getPropertyType();
    }
    public Class getDeclaringClass()
    {
        return this.rootObjectType;
    }


    /**
     * Sets aValue for anObject
     */
    public void doSet(Object anObject, Object aValue) throws MetadataException
    {
        if(anObject == null) return;
        Method m = getPropertyDescriptor().getWriteMethod();
        Object[] args = {aValue};
        if (m != null)
        {
            try
            {
                /**
                 * MBAIRD: it is safe to call getParameterTypes()[0] because this is
                 * the "set" method and it needs to take one parameter only.
                 * we need to be able to set values to null. We can only set something to null if
                 * the type is not a primitive (assignable from Object).
                 */
                if ((aValue != null) || !m.getParameterTypes()[0].isPrimitive())
                {
                    m.invoke(anObject, args);
                }
            }
            catch (Throwable e)
            {
                logProblem(anObject, aValue, e.getClass().getName());
                throw new MetadataException("Error invoking method:" + m.getName() + " in object:" + anObject.getClass().getName(), e);
            }
        }
        else
        {
            logProblem(anObject, aValue, "set: getWriteMethod returned null");
            throw new MetadataException(
                    "Can't get WriteMethod for property:" + getPropertyDescriptor().getName() + " in object:" + anObject.getClass().getName());
        }
    }

    /**
     * Get the Value from anObject
     */
    public Object doGet(Object anObject) throws MetadataException
    {
        if(anObject == null) return null;
        Method m = getPropertyDescriptor().getReadMethod();
        Object[] args = null;

        if (m != null)
        {
            try
            {
                Object result = m.invoke(anObject, args);
                return result;
            }
            catch (Throwable e)
            {
                logProblem(anObject, null, e.getClass().getName());
                throw new MetadataException("Error invoking method:" + m.getName() + " in object:" + anObject.getClass().getName(), e);
            }
        }
        else
        {
            logProblem(anObject, null, "get: getReadMethod returned null");
            throw new MetadataException(
                    "Can't get ReadMethod for property:" + getPropertyDescriptor().getName() + " in object:" + anObject.getClass().getName());
        }
    }

    /**
     * Get the PropertyDescriptor for aClass and aPropertyName
     */
    protected static PropertyDescriptor findPropertyDescriptor(Class aClass, String aPropertyName)
    {
        BeanInfo info;
        PropertyDescriptor[] pd;
        PropertyDescriptor descriptor = null;

        try
        {
            info = Introspector.getBeanInfo(aClass);
            pd = info.getPropertyDescriptors();
            for (int i = 0; i < pd.length; i++)
            {
                if (pd[i].getName().equals(aPropertyName))
                {
                    descriptor = pd[i];
                    break;
                }
            }
            if (descriptor == null)
            {
                /*
				 * Daren Drummond: 	Throw here so we are consistent
				 * 					with PersistentFieldDefaultImpl.
				 */
                throw new MetadataException("Can't find property " + aPropertyName + " in " + aClass.getName());
            }
            return descriptor;
        }
        catch (IntrospectionException ex)
        {
            /*
			 * Daren Drummond: 	Throw here so we are consistent
			 * 					with PersistentFieldDefaultImpl.
			 */
            throw new MetadataException("Can't find property " + aPropertyName + " in " + aClass.getName(), ex);
        }

    }

    /**
     * Returns the PropertyDescriptor.
     * @return java.beans.PropertyDescriptor
     */
    protected PropertyDescriptor getPropertyDescriptor()
    {
        if (propertyDescriptor == null)
        {
            setPropertyDescriptor(findPropertyDescriptor(getDeclaringClass(), getName()));
        }
        return propertyDescriptor;
    }

    /**
     * Set PropertyDescriptor.
     * @param newPropertyDescriptor java.beans.PropertyDescriptor
     */
    protected void setPropertyDescriptor(PropertyDescriptor newPropertyDescriptor)
    {
        propertyDescriptor = newPropertyDescriptor;
    }

    /**
     * This implementation returns always 'false'.
     * @see AbstractPersistentField#makeAccessible()
     */
    public boolean makeAccessible()
    {
        return false;
    }

    /**
     * Always returns 'false'.
     * @see PersistentField#usesAccessorsAndMutators
     */
    public boolean usesAccessorsAndMutators()
    {
        return true;
    }

    /**
     * Let's give the user some hints as to what could be wrong.
     */
    protected void logProblem(Object anObject, Object aValue, String msg)
    {
        Logger logger = LoggerFactory.getDefaultLogger();
        logger.error("Error in [PersistentFieldPropertyImpl], " + msg);
        logger.error("Declaring class [" + getDeclaringClass().getName() + "]");
        logger.error("Property Name [" + getName() + "]");
        logger.error("Property Type [" + getPropertyDescriptor().getPropertyType().getName() + "]");

        if (anObject != null)
        {
            logger.error("anObject was class [" + anObject.getClass().getName() + "]");
        }
        else
        {
            logger.error("anObject was null");
        }
        if (aValue != null)
        {
            logger.error("aValue was class [" + aValue.getClass().getName() + "]");
        }
        else
        {
            logger.error("aValue was null");
        }
    }
}
